from __future__ import annotations

import itertools
from typing import TYPE_CHECKING, cast

from pdm.models.repositories import BaseRepository
from pdm.models.requirements import strip_extras
from pdm.resolver.providers import BaseProvider, EagerUpdateProvider, ReusePinProvider
from pdm.resolver.python import find_python_matches
from pdm.utils import is_url
from unearth.utils import LazySequence

from pdm_conda.models.repositories import CondaRepository
from pdm_conda.models.requirements import (
    CondaRequirement,
    as_conda_requirement,
    parse_requirement,
)

if TYPE_CHECKING:
    from typing import Callable, Iterable, Iterator, Mapping, Sequence

    from pdm.models.candidates import Candidate
    from pdm.models.requirements import Requirement
    from pdm.resolver.providers import Comparable
    from resolvelib.resolvers import RequirementInformation


class CondaBaseProvider(BaseProvider):
    def __init__(
        self,
        repository: BaseRepository,
        allow_prereleases: bool | None = None,
        overrides: dict[str, str] | None = None,
        direct_minimal_versions: bool = False,
    ) -> None:
        super().__init__(repository, allow_prereleases, overrides, direct_minimal_versions)
        self._overrides_requirements: dict | None = None

    @property
    def overrides_requirements(self) -> dict[str, Requirement]:
        """
        Identifier and requirement mapping for overrides
        :return: mapping
        """
        if self._overrides_requirements is None:
            self._overrides_requirements = dict()
            if self.overrides:
                for identifier, requested in self.overrides.items():
                    if is_url(requested):
                        requirement = parse_requirement(f"{identifier} @ {requested}")
                    else:
                        # first parse as conda to ensure no version error
                        requirement = cast(CondaRequirement, parse_requirement(f"conda:{identifier} {requested}"))
                        if not isinstance(self.repository, CondaRepository) or not self.repository.is_conda_managed(
                            requirement,
                        ):
                            requirement.name = identifier
                            requirement = parse_requirement(requirement.as_line())
                    self._overrides_requirements[self.identify(requirement)] = requirement

        return self._overrides_requirements

    def get_preference(
        self,
        identifier: str,
        resolutions: dict[str, Candidate],
        candidates: dict[str, Iterator[Candidate]],
        information: dict[str, Iterator[RequirementInformation]],
        backtrack_causes: Sequence[RequirementInformation],
    ) -> tuple[Comparable, ...]:
        preference = super().get_preference(identifier, resolutions, candidates, information, backtrack_causes)
        return (
            *preference[:3],
            not any(isinstance(i.requirement, CondaRequirement) for i in information[identifier]),
            *preference[3:],
        )

    def find_matches(
        self,
        identifier: str,
        requirements: Mapping[str, Iterator[Requirement]],
        incompatibilities: Mapping[str, Iterator[Candidate]],
    ) -> Callable[[], Iterator[Candidate]]:
        def matches_gen() -> Iterator[Candidate]:
            incompat = list(incompatibilities[identifier])
            if identifier == "python":
                candidates = find_python_matches(identifier, requirements)
                return (c for c in candidates if c not in incompat)
            elif identifier in self.overrides:
                return iter(self.get_override_candidates(identifier))
            reqs_iter = requirements[identifier]
            bare_name, extras = strip_extras(identifier)
            if extras and bare_name in requirements:
                # We should consider the requirements for both foo and foo[extra]
                reqs_iter = itertools.chain(reqs_iter, requirements[bare_name])
            reqs = sorted(reqs_iter, key=self.requirement_preference)
            # iterates over requirements
            candidates = []
            for req in reqs:
                candidates = self._find_candidates(req)
                candidates = LazySequence(
                    can for can in candidates if can not in incompat and all(self.is_satisfied_by(r, can) for r in reqs)
                )
                if candidates:
                    break
            return iter(candidates)

        return matches_gen

    def get_requirement_from_overrides(self, requirement: Requirement) -> Requirement:
        _req = self.overrides_requirements.get(self.identify(requirement), requirement)
        if isinstance(requirement, CondaRequirement):
            _req = as_conda_requirement(_req)
        return _req

    def get_override_candidates(self, identifier: str) -> Iterable[Candidate]:
        return self._find_candidates(self.overrides_requirements[identifier])


class CondaReusePinProvider(ReusePinProvider, CondaBaseProvider):
    pass


class CondaEagerUpdateProvider(EagerUpdateProvider, CondaBaseProvider):
    pass



from __future__ import annotations

from typing import TYPE_CHECKING, cast

from pdm.cli.commands.remove import Command as BaseCommand

from pdm_conda.cli.utils import remove_quotes
from pdm_conda.models.requirements import parse_requirement
from pdm_conda.project import CondaProject

if TYPE_CHECKING:
    import argparse

    from pdm_conda.project import Project


class Command(BaseCommand):
    description = BaseCommand.__doc__
    name = "remove"

    def handle(self, project: Project, options: argparse.Namespace) -> None:
        if options.group is None:
            options.group = "dev" if options.dev else "default"

        project = cast(CondaProject, project)
        conda_dependencies = project.get_conda_pyproject_dependencies(options.group, options.dev)
        dependencies, _ = project.use_pyproject_dependencies(options.group, options.dev)
        _dependencies = [parse_requirement(d).conda_name for d in dependencies]
        # add conda dependencies to common dependencies if going to remove it
        for i, pkg in enumerate(options.packages):
            # parse it as conda, if found add it to dependencies
            conda_package = f"conda:{remove_quotes(pkg)}"
            package = parse_requirement(conda_package)
            idx = None
            for dep_idx, dep in enumerate(conda_dependencies):
                dep = parse_requirement(f"conda:{dep}")
                if package.name == dep.name:
                    if package.name not in _dependencies:
                        dependencies.append(conda_package)
                    options.packages[i] = conda_package
                    idx = dep_idx
                    break
            if idx is not None:
                conda_dependencies.pop(idx)
        super().handle(project, options)



from __future__ import annotations

from typing import TYPE_CHECKING, cast

from pdm.cli.commands.update import Command as BaseCommand
from pdm.models.specifiers import get_specifier

from pdm_conda.project import CondaProject

if TYPE_CHECKING:
    import argparse

    from pdm_conda.project import Project


class Command(BaseCommand):
    description = BaseCommand.__doc__
    name = "update"

    def handle(self, project: Project, options: argparse.Namespace) -> None:
        super().handle(project=project, options=options)
        project = cast(CondaProject, project)
        if project.conda_config.use_custom_behaviour:
            if not options.dry_run and options.save_strategy:
                groups = set(project.iter_groups(dev=False))
                dev_groups = set(project.iter_groups(dev=True)) - groups
                num_groups = len(groups) + len(dev_groups)
                candidates = project.locked_repository.all_candidates
                for i, group in enumerate(groups | dev_groups):
                    requirements = dict()
                    for identifier, _ in project.get_dependencies(group).items():
                        req = candidates.get(identifier).req

                        if req.is_named:
                            version = str(req.specifier).removeprefix("==")
                            if options.save_strategy == "minimum":
                                req.specifier = get_specifier(f">={version}")
                            elif options.save_strategy == "compatible":
                                req.specifier = get_specifier(f"~={version}")
                        requirements[identifier] = req
                    project.add_dependencies(requirements, dev=i >= len(groups), show_message=i == num_groups - 1)



import functools
from copy import copy

from pdm.environments import PythonEnvironment

from pdm_conda.project import CondaProject

_patched = False


def wrap_init(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        res = func(self, *args, **kwargs)
        if isinstance(self.project, CondaProject):
            self.project = copy(self.project)
            self.project.environment = self
        return res

    return wrapper


if not _patched:
    setattr(PythonEnvironment, "__init__", wrap_init(PythonEnvironment.__init__))
    _patched = True
